@@ -27,8 +27,10 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "../../src/client/registry.h"
 #include "../../src/client/output.h"
 #include "../../src/client/seat.h"
+#include "../../src/client/server_decoration.h"
 #include "../../src/client/shell.h"
 #include "../../src/client/subcompositor.h"
+#include "../../src/client/xdgshell.h"
 #include "../../src/server/compositor_interface.h"
 #include "../../src/server/datadevicemanager_interface.h"
 #include "../../src/server/display.h"
@@ -38,13 +40,20 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "../../src/server/shell_interface.h"
 #include "../../src/server/blur_interface.h"
 #include "../../src/server/contrast_interface.h"
+#include "../../src/server/server_decoration_interface.h"
 #include "../../src/server/slide_interface.h"
 #include "../../src/server/subcompositor_interface.h"
 #include "../../src/server/outputmanagement_interface.h"
 #include "../../src/server/outputdevice_interface.h"
+#include "../../src/server/textinput_interface.h"
+#include "../../src/server/xdgshell_interface.h"
 // Wayland
 #include <wayland-client-protocol.h>
 #include <wayland-dpms-client-protocol.h>
+#include <wayland-server-decoration-client-protocol.h>
+#include <wayland-text-input-v0-client-protocol.h>
+#include <wayland-text-input-v2-client-protocol.h>
+#include <wayland-xdg-shell-v5-client-protocol.h>
 
 class TestWaylandRegistry : public QObject
 {
@@ -67,6 +76,10 @@ private Q_SLOTS:
     void testBindContrastManager();
     void testBindSlideManager();
     void testBindDpmsManager();
+    void testBindServerSideDecorationManager();
+    void testBindTextInputManagerUnstableV0();
+    void testBindTextInputManagerUnstableV2();
+    void testBindXdgShellUnstableV5();
     void testGlobalSync();
     void testGlobalSyncThreaded();
     void testRemoval();
@@ -84,6 +97,10 @@ private:
     KWayland::Server::SubCompositorInterface *m_subcompositor;
     KWayland::Server::DataDeviceManagerInterface *m_dataDeviceManager;
     KWayland::Server::OutputManagementInterface *m_outputManagement;
+    KWayland::Server::ServerSideDecorationManagerInterface *m_serverSideDecorationManager;
+    KWayland::Server::TextInputManagerInterface *m_textInputManagerV0;
+    KWayland::Server::TextInputManagerInterface *m_textInputManagerV2;
+    KWayland::Server::XdgShellInterface *m_xdgShellUnstableV5;
 };
 
 static const QString s_socketName = QStringLiteral("kwin-test-wayland-registry-0");
@@ -99,6 +116,10 @@ TestWaylandRegistry::TestWaylandRegistry(QObject *parent)
     , m_subcompositor(nullptr)
     , m_dataDeviceManager(nullptr)
     , m_outputManagement(nullptr)
+    , m_serverSideDecorationManager(nullptr)
+    , m_textInputManagerV0(nullptr)
+    , m_textInputManagerV2(nullptr)
+    , m_xdgShellUnstableV5(nullptr)
 {
 }
 
@@ -129,6 +150,17 @@ void TestWaylandRegistry::init()
     m_display->createContrastManager(this)->create();
     m_display->createSlideManager(this)->create();
     m_display->createDpmsManager()->create();
+    m_serverSideDecorationManager = m_display->createServerSideDecorationManager();
+    m_serverSideDecorationManager->create();
+    m_textInputManagerV0 = m_display->createTextInputManager(KWayland::Server::TextInputInterfaceVersion::UnstableV0);
+    QCOMPARE(m_textInputManagerV0->interfaceVersion(), KWayland::Server::TextInputInterfaceVersion::UnstableV0);
+    m_textInputManagerV0->create();
+    m_textInputManagerV2 = m_display->createTextInputManager(KWayland::Server::TextInputInterfaceVersion::UnstableV2);
+    QCOMPARE(m_textInputManagerV2->interfaceVersion(), KWayland::Server::TextInputInterfaceVersion::UnstableV2);
+    m_textInputManagerV2->create();
+    m_xdgShellUnstableV5 = m_display->createXdgShell(KWayland::Server::XdgShellInterfaceVersion::UnstableV5);
+    m_xdgShellUnstableV5->create();
+    QCOMPARE(m_xdgShellUnstableV5->interfaceVersion(), KWayland::Server::XdgShellInterfaceVersion::UnstableV5);
 }
 
 void TestWaylandRegistry::cleanup()
@@ -251,6 +283,26 @@ void TestWaylandRegistry::testBindDpmsManager()
     TEST_BIND(KWayland::Client::Registry::Interface::Dpms, SIGNAL(dpmsAnnounced(quint32,quint32)), bindDpmsManager, org_kde_kwin_dpms_manager_destroy)
 }
 
+void TestWaylandRegistry::testBindServerSideDecorationManager()
+{
+    TEST_BIND(KWayland::Client::Registry::Interface::ServerSideDecorationManager, SIGNAL(serverSideDecorationManagerAnnounced(quint32,quint32)), bindServerSideDecorationManager, org_kde_kwin_server_decoration_manager_destroy)
+}
+
+void TestWaylandRegistry::testBindTextInputManagerUnstableV0()
+{
+    TEST_BIND(KWayland::Client::Registry::Interface::TextInputManagerUnstableV0, SIGNAL(textInputManagerUnstableV0Announced(quint32,quint32)), bindTextInputManagerUnstableV0, wl_text_input_manager_destroy)
+}
+
+void TestWaylandRegistry::testBindTextInputManagerUnstableV2()
+{
+    TEST_BIND(KWayland::Client::Registry::Interface::TextInputManagerUnstableV2, SIGNAL(textInputManagerUnstableV2Announced(quint32,quint32)), bindTextInputManagerUnstableV2, zwp_text_input_manager_v2_destroy)
+}
+
+void TestWaylandRegistry::testBindXdgShellUnstableV5()
+{
+    TEST_BIND(KWayland::Client::Registry::Interface::XdgShellUnstableV5, SIGNAL(xdgShellUnstableV5Announced(quint32,quint32)), bindXdgShellUnstableV5, xdg_shell_destroy)
+}
+
 #undef TEST_BIND
 
 void TestWaylandRegistry::testRemoval()
@@ -284,6 +336,8 @@ void TestWaylandRegistry::testRemoval()
     QVERIFY(subCompositorAnnouncedSpy.isValid());
     QSignalSpy outputManagementAnnouncedSpy(&registry, SIGNAL(outputManagementAnnounced(quint32,quint32)));
     QVERIFY(outputManagementAnnouncedSpy.isValid());
+    QSignalSpy serverSideDecorationManagerAnnouncedSpy(&registry, &Registry::serverSideDecorationManagerAnnounced);
+    QVERIFY(serverSideDecorationManagerAnnouncedSpy.isValid());
 
     QVERIFY(!registry.isValid());
     registry.create(connection.display());
@@ -297,6 +351,7 @@ void TestWaylandRegistry::testRemoval()
     QVERIFY(!seatAnnouncedSpy.isEmpty());
     QVERIFY(!subCompositorAnnouncedSpy.isEmpty());
     QVERIFY(!outputManagementAnnouncedSpy.isEmpty());
+    QVERIFY(!serverSideDecorationManagerAnnouncedSpy.isEmpty());
 
     QVERIFY(registry.hasInterface(KWayland::Client::Registry::Interface::Compositor));
     QVERIFY(registry.hasInterface(KWayland::Client::Registry::Interface::Output));
@@ -307,6 +362,7 @@ void TestWaylandRegistry::testRemoval()
     QVERIFY(registry.hasInterface(KWayland::Client::Registry::Interface::SubCompositor));
     QVERIFY(!registry.hasInterface(KWayland::Client::Registry::Interface::FullscreenShell));
     QVERIFY(registry.hasInterface(KWayland::Client::Registry::Interface::OutputManagement));
+    QVERIFY(registry.hasInterface(KWayland::Client::Registry::Interface::ServerSideDecorationManager));
 
     QVERIFY(!registry.interfaces(KWayland::Client::Registry::Interface::Compositor).isEmpty());
     QVERIFY(!registry.interfaces(KWayland::Client::Registry::Interface::Output).isEmpty());
@@ -317,6 +373,7 @@ void TestWaylandRegistry::testRemoval()
     QVERIFY(!registry.interfaces(KWayland::Client::Registry::Interface::SubCompositor).isEmpty());
     QVERIFY(registry.interfaces(KWayland::Client::Registry::Interface::FullscreenShell).isEmpty());
     QVERIFY(!registry.interfaces(KWayland::Client::Registry::Interface::OutputManagement).isEmpty());
+    QVERIFY(!registry.interfaces(KWayland::Client::Registry::Interface::ServerSideDecorationManager).isEmpty());
 
     QSignalSpy seatRemovedSpy(&registry, SIGNAL(seatRemoved(quint32)));
     QVERIFY(seatRemovedSpy.isValid());
@@ -326,6 +383,7 @@ void TestWaylandRegistry::testRemoval()
     Output *output = registry.createOutput(registry.interface(Registry::Interface::Output).name, registry.interface(Registry::Interface::Output).version, &registry);
     Compositor *compositor = registry.createCompositor(registry.interface(Registry::Interface::Compositor).name, registry.interface(Registry::Interface::Compositor).version, &registry);
     SubCompositor *subcompositor = registry.createSubCompositor(registry.interface(Registry::Interface::SubCompositor).name, registry.interface(Registry::Interface::SubCompositor).version, &registry);
+    ServerSideDecorationManager *serverSideDeco = registry.createServerSideDecorationManager(registry.interface(Registry::Interface::ServerSideDecorationManager).name, registry.interface(Registry::Interface::ServerSideDecorationManager).version, &registry);
     connection.flush();
     m_display->dispatchEvents();
     QSignalSpy seatObjectRemovedSpy(seat, &Seat::removed);
@@ -404,6 +462,18 @@ void TestWaylandRegistry::testRemoval()
     QVERIFY(!registry.hasInterface(KWayland::Client::Registry::Interface::OutputManagement));
     QVERIFY(registry.interfaces(KWayland::Client::Registry::Interface::OutputManagement).isEmpty());
 
+    QSignalSpy serverSideDecoManagerRemovedSpy(&registry, &Registry::serverSideDecorationManagerRemoved);
+    QVERIFY(serverSideDecoManagerRemovedSpy.isValid());
+    QSignalSpy serverSideDecoManagerObjectRemovedSpy(serverSideDeco, &ServerSideDecorationManager::removed);
+    QVERIFY(serverSideDecoManagerObjectRemovedSpy.isValid());
+
+    delete m_serverSideDecorationManager;
+    QVERIFY(serverSideDecoManagerRemovedSpy.wait());
+    QCOMPARE(serverSideDecoManagerRemovedSpy.first().first(), serverSideDecorationManagerAnnouncedSpy.first().first());
+    QVERIFY(!registry.hasInterface(KWayland::Client::Registry::Interface::ServerSideDecorationManager));
+    QVERIFY(registry.interfaces(KWayland::Client::Registry::Interface::ServerSideDecorationManager).isEmpty());
+    QCOMPARE(serverSideDecoManagerObjectRemovedSpy.count(), 1);
+
     // cannot test shmRemoved as there is no functionality for it
 
     // verify everything has been removed only once
@@ -412,6 +482,7 @@ void TestWaylandRegistry::testRemoval()
     QCOMPARE(outputObjectRemovedSpy.count(), 1);
     QCOMPARE(compositorObjectRemovedSpy.count(), 1);
     QCOMPARE(subcompositorObjectRemovedSpy.count(), 1);
+    QCOMPARE(serverSideDecoManagerObjectRemovedSpy.count(), 1);
 }
 
 void TestWaylandRegistry::testDestroy()
