@@ -79,6 +79,7 @@ private Q_SLOTS:
     void testSelection();
     void testTouch();
     void testDisconnect();
+    void testPointerEnterOnUnboundSurface();
     // TODO: add test for keymap
 
 private:
@@ -510,7 +511,11 @@ void TestWaylandSeat::testPointer()
     serverPointer = m_seatInterface->focusedPointer();
     QVERIFY(serverPointer);
 
+    QSignalSpy entered2Spy(p, &Pointer::entered);
+    QVERIFY(entered2Spy.wait());
+    QCOMPARE(p->enteredSurface(), s);
     delete s;
+    QVERIFY(!p->enteredSurface());
     wl_display_flush(m_connection->display());
     QVERIFY(focusedPointerChangedSpy.wait());
     QCOMPARE(focusedPointerChangedSpy.count(), 10);
@@ -1409,11 +1414,10 @@ void TestWaylandSeat::testSelection()
     m_seatInterface->setFocusedKeyboardSurface(serverSurface);
     QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
     QVERIFY(!m_seatInterface->focusedKeyboard());
-    serverSurface->client()->flush();
-    QCoreApplication::processEvents();
-    QCoreApplication::processEvents();
+    QVERIFY(selectionClearedSpy.wait());
     QVERIFY(selectionSpy.isEmpty());
-    QVERIFY(selectionClearedSpy.isEmpty());
+    QVERIFY(!selectionClearedSpy.isEmpty());
+    selectionClearedSpy.clear();
     QVERIFY(!m_seatInterface->selection());
 
     // now let's try to set a selection - we have keyboard focus, so it should be sent to us
@@ -1425,7 +1429,6 @@ void TestWaylandSeat::testSelection()
     QCOMPARE(selectionSpy.count(), 1);
     auto ddi = m_seatInterface->selection();
     QVERIFY(ddi);
-    QVERIFY(selectionClearedSpy.isEmpty());
     auto df = selectionSpy.first().first().value<DataOffer*>();
     QCOMPARE(df->offeredMimeTypes().count(), 1);
     QCOMPARE(df->offeredMimeTypes().first().name(), QStringLiteral("text/plain"));
@@ -1758,5 +1761,40 @@ void TestWaylandSeat::testDisconnect()
     m_queue->destroy();
 }
 
+void TestWaylandSeat::testPointerEnterOnUnboundSurface()
+{
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+
+    // create the things we need
+    m_seatInterface->setHasKeyboard(true);
+    m_seatInterface->setHasPointer(true);
+    m_seatInterface->setHasTouch(true);
+    QSignalSpy pointerChangedSpy(m_seat, &Seat::hasPointerChanged);
+    QVERIFY(pointerChangedSpy.isValid());
+    QVERIFY(pointerChangedSpy.wait());
+
+    // create pointer and Surface
+    QScopedPointer<Pointer> pointer(m_seat->createPointer());
+    QVERIFY(!pointer.isNull());
+    // create the surface
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, &CompositorInterface::surfaceCreated);
+    QVERIFY(surfaceCreatedSpy.isValid());
+    QScopedPointer<Surface> s(m_compositor->createSurface());
+    QVERIFY(surfaceCreatedSpy.wait());
+    SurfaceInterface *serverSurface = surfaceCreatedSpy.first().first().value<KWayland::Server::SurfaceInterface*>();
+    QVERIFY(serverSurface);
+
+    // unbind the surface again
+    QSignalSpy surfaceUnboundSpy(serverSurface, &SurfaceInterface::unbound);
+    QVERIFY(surfaceUnboundSpy.isValid());
+    s.reset();
+    QVERIFY(surfaceUnboundSpy.wait());
+    QSignalSpy clientErrorSpy(m_connection, &ConnectionThread::errorOccurred);
+    QVERIFY(clientErrorSpy.isValid());
+    m_seatInterface->setFocusedPointerSurface(serverSurface);
+    QVERIFY(!clientErrorSpy.wait(100));
+}
+
 QTEST_GUILESS_MAIN(TestWaylandSeat)
 #include "test_wayland_seat.moc"
