@@ -21,10 +21,13 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "event_queue.h"
 #include "seat.h"
 #include "wayland_pointer_p.h"
-
+#include <QPointF>
 #include <QSizeF>
 
+#include <config-kwayland.h>
+#if HAVE_LINUX_INPUT_H
 #include <linux/input.h>
+#endif
 
 #include <wayland-fake-input-client-protocol.h>
 
@@ -99,6 +102,7 @@ void FakeInput::requestPointerMove(const QSizeF &delta)
 
 void FakeInput::Private::sendPointerButtonState(Qt::MouseButton button, quint32 state)
 {
+#if HAVE_LINUX_INPUT_H
     Q_ASSERT(manager.isValid());
     uint32_t b = 0;
     switch (button) {
@@ -117,6 +121,7 @@ void FakeInput::Private::sendPointerButtonState(Qt::MouseButton button, quint32
         return;
     }
     org_kde_kwin_fake_input_button(manager, b, state);
+#endif
 }
 
 void FakeInput::requestPointerButtonPress(Qt::MouseButton button)
@@ -171,6 +176,36 @@ void FakeInput::requestPointerAxis(Qt::Orientation axis, qreal delta)
     org_kde_kwin_fake_input_axis(d->manager, a, wl_fixed_from_double(delta));
 }
 
+void FakeInput::requestTouchDown(quint32 id, const QPointF &pos)
+{
+    Q_ASSERT(d->manager.isValid());
+    org_kde_kwin_fake_input_touch_down(d->manager, id, wl_fixed_from_double(pos.x()), wl_fixed_from_double(pos.y()));
+}
+
+void FakeInput::requestTouchMotion(quint32 id, const QPointF &pos)
+{
+    Q_ASSERT(d->manager.isValid());
+    org_kde_kwin_fake_input_touch_motion(d->manager, id, wl_fixed_from_double(pos.x()), wl_fixed_from_double(pos.y()));
+}
+
+void FakeInput::requestTouchUp(quint32 id)
+{
+    Q_ASSERT(d->manager.isValid());
+    org_kde_kwin_fake_input_touch_up(d->manager, id);
+}
+
+void FakeInput::requestTouchCancel()
+{
+    Q_ASSERT(d->manager.isValid());
+    org_kde_kwin_fake_input_touch_cancel(d->manager);
+}
+
+void FakeInput::requestTouchFrame()
+{
+    Q_ASSERT(d->manager.isValid());
+    org_kde_kwin_fake_input_touch_frame(d->manager);
+}
+
 FakeInput::operator org_kde_kwin_fake_input*() const
 {
     return d->manager;
