@@ -55,6 +55,79 @@ private:
     PlasmaWindowManagement *q;
 };
 
+class PlasmaWindow::Private
+{
+public:
+    Private(org_kde_plasma_window *window, quint32 internalId, PlasmaWindow *q);
+    WaylandPointer<org_kde_plasma_window, org_kde_plasma_window_destroy> window;
+    quint32 internalId;
+    QString title;
+    QString appId;
+    quint32 desktop = 0;
+    bool active = false;
+    bool minimized = false;
+    bool maximized = false;
+    bool fullscreen = false;
+    bool keepAbove = false;
+    bool keepBelow = false;
+    bool onAllDesktops = false;
+    bool demandsAttention = false;
+    bool closeable = false;
+    bool minimizeable = false;
+    bool maximizeable = false;
+    bool fullscreenable = false;
+    bool skipTaskbar = false;
+    bool shadeable = false;
+    bool shaded = false;
+    bool movable = false;
+    bool resizable = false;
+    bool virtualDesktopChangeable = false;
+    QIcon icon;
+    PlasmaWindowManagement *wm = nullptr;
+    bool unmapped = false;
+    QPointer<PlasmaWindow> parentWindow;
+    QMetaObject::Connection parentWindowUnmappedConnection;
+    QRect geometry;
+
+private:
+    static void titleChangedCallback(void *data, org_kde_plasma_window *window, const char *title);
+    static void appIdChangedCallback(void *data, org_kde_plasma_window *window, const char *app_id);
+    static void stateChangedCallback(void *data, org_kde_plasma_window *window, uint32_t state);
+    static void virtualDesktopChangedCallback(void *data, org_kde_plasma_window *window, int32_t number);
+    static void themedIconNameChangedCallback(void *data, org_kde_plasma_window *window, const char *name);
+    static void unmappedCallback(void *data, org_kde_plasma_window *window);
+    static void initialStateCallback(void *data, org_kde_plasma_window *window);
+    static void parentWindowCallback(void *data, org_kde_plasma_window *window, org_kde_plasma_window *parent);
+    static void windowGeometryCallback(void *data, org_kde_plasma_window *window, int32_t x, int32_t y, uint32_t width, uint32_t height);
+    void setActive(bool set);
+    void setMinimized(bool set);
+    void setMaximized(bool set);
+    void setFullscreen(bool set);
+    void setKeepAbove(bool set);
+    void setKeepBelow(bool set);
+    void setOnAllDesktops(bool set);
+    void setDemandsAttention(bool set);
+    void setCloseable(bool set);
+    void setMinimizeable(bool set);
+    void setMaximizeable(bool set);
+    void setFullscreenable(bool set);
+    void setSkipTaskbar(bool skip);
+    void setShadeable(bool set);
+    void setShaded(bool set);
+    void setMovable(bool set);
+    void setResizable(bool set);
+    void setVirtualDesktopChangeable(bool set);
+    void setParentWindow(PlasmaWindow *parentWindow);
+
+    static Private *cast(void *data) {
+        return reinterpret_cast<Private*>(data);
+    }
+
+    PlasmaWindow *q;
+
+    static struct org_kde_plasma_window_listener s_listener;
+};
+
 PlasmaWindowManagement::Private::Private(PlasmaWindowManagement *q)
     : q(q)
 {
@@ -121,10 +194,23 @@ void PlasmaWindowManagement::Private::windowCreated(org_kde_plasma_window *id, q
         queue->addProxy(id);
     }
     PlasmaWindow *window = new PlasmaWindow(q, id, internalId);
+    window->d->wm = q;
     windows << window;
     QObject::connect(window, &QObject::destroyed, q,
         [this, window] {
             windows.removeAll(window);
+            if (activeWindow == window) {
+                activeWindow = nullptr;
+                emit q->activeWindowChanged();
+            }
+        }
+    );
+    QObject::connect(window, &PlasmaWindow::unmapped, q,
+        [this, window] {
+            if (activeWindow == window) {
+                activeWindow = nullptr;
+                emit q->activeWindowChanged();
+            }
         }
     );
     QObject::connect(window, &PlasmaWindow::activeChanged, q,
@@ -143,7 +229,6 @@ void PlasmaWindowManagement::Private::windowCreated(org_kde_plasma_window *id, q
             }
         }
     );
-    emit q->windowCreated(window);
 }
 
 PlasmaWindowManagement::PlasmaWindowManagement(QObject *parent)
@@ -240,70 +325,72 @@ PlasmaWindowModel *PlasmaWindowManagement::createWindowModel()
     return new PlasmaWindowModel(this);
 }
 
-
-class PlasmaWindow::Private
-{
-public:
-    Private(org_kde_plasma_window *window, quint32 internalId, PlasmaWindow *q);
-    WaylandPointer<org_kde_plasma_window, org_kde_plasma_window_destroy> window;
-    quint32 internalId;
-    QString title;
-    QString appId;
-    quint32 desktop = 0;
-    bool active = false;
-    bool minimized = false;
-    bool maximized = false;
-    bool fullscreen = false;
-    bool keepAbove = false;
-    bool keepBelow = false;
-    bool onAllDesktops = false;
-    bool demandsAttention = false;
-    bool closeable = false;
-    bool minimizeable = false;
-    bool maximizeable = false;
-    bool fullscreenable = false;
-    bool skipTaskbar = false;
-    QIcon icon;
-
-private:
-    static void titleChangedCallback(void *data, org_kde_plasma_window *window, const char *title);
-    static void appIdChangedCallback(void *data, org_kde_plasma_window *window, const char *app_id);
-    static void stateChangedCallback(void *data, org_kde_plasma_window *window, uint32_t state);
-    static void virtualDesktopChangedCallback(void *data, org_kde_plasma_window *window, int32_t number);
-    static void themedIconNameChangedCallback(void *data, org_kde_plasma_window *window, const char *name);
-    static void unmappedCallback(void *data, org_kde_plasma_window *window);
-    void setActive(bool set);
-    void setMinimized(bool set);
-    void setMaximized(bool set);
-    void setFullscreen(bool set);
-    void setKeepAbove(bool set);
-    void setKeepBelow(bool set);
-    void setOnAllDesktops(bool set);
-    void setDemandsAttention(bool set);
-    void setCloseable(bool set);
-    void setMinimizeable(bool set);
-    void setMaximizeable(bool set);
-    void setFullscreenable(bool set);
-    void setSkipTaskbar(bool skip);
-
-    static Private *cast(void *data) {
-        return reinterpret_cast<Private*>(data);
-    }
-
-    PlasmaWindow *q;
-
-    static struct org_kde_plasma_window_listener s_listener;
-};
-
 org_kde_plasma_window_listener PlasmaWindow::Private::s_listener = {
     titleChangedCallback,
     appIdChangedCallback,
     stateChangedCallback,
     virtualDesktopChangedCallback,
     themedIconNameChangedCallback,
-    unmappedCallback
+    unmappedCallback,
+    initialStateCallback,
+    parentWindowCallback,
+    windowGeometryCallback
 };
 
+void PlasmaWindow::Private::parentWindowCallback(void *data, org_kde_plasma_window *window, org_kde_plasma_window *parent)
+{
+    Q_UNUSED(window)
+    Private *p = cast(data);
+    const auto windows = p->wm->windows();
+    auto it = std::find_if(windows.begin(), windows.end(),
+        [parent] (const PlasmaWindow *w) {
+            return *w == parent;
+        }
+    );
+    p->setParentWindow(it != windows.end() ? *it : nullptr);
+}
+
+void PlasmaWindow::Private::windowGeometryCallback(void *data, org_kde_plasma_window *window, int32_t x, int32_t y, uint32_t width, uint32_t height)
+{
+    Q_UNUSED(window)
+    Private *p = cast(data);
+    QRect geo(x, y, width, height);
+    if (geo == p->geometry) {
+        return;
+    }
+    p->geometry = geo;
+    emit p->q->geometryChanged();
+}
+
+void PlasmaWindow::Private::setParentWindow(PlasmaWindow *parent)
+{
+    const auto old = parentWindow;
+    QObject::disconnect(parentWindowUnmappedConnection);
+    if (parent && !parent->d->unmapped) {
+        parentWindow = QPointer<PlasmaWindow>(parent);
+        parentWindowUnmappedConnection = QObject::connect(parent, &PlasmaWindow::unmapped, q,
+            [this] {
+                setParentWindow(nullptr);
+            }
+        );
+    } else {
+        parentWindow = QPointer<PlasmaWindow>();
+        parentWindowUnmappedConnection = QMetaObject::Connection();
+    }
+    if (parentWindow.data() != old.data()) {
+        emit q->parentWindowChanged();
+    }
+}
+
+void PlasmaWindow::Private::initialStateCallback(void *data, org_kde_plasma_window *window)
+{
+    Q_UNUSED(window)
+    Private *p = cast(data);
+    if (!p->unmapped) {
+        emit p->wm->windowCreated(p->q);
+    }
+}
+
 void PlasmaWindow::Private::titleChangedCallback(void *data, org_kde_plasma_window *window, const char *title)
 {
     Q_UNUSED(window)
@@ -332,7 +419,7 @@ void PlasmaWindow::Private::virtualDesktopChangedCallback(void *data, org_kde_pl
 {
     Q_UNUSED(window)
     Private *p = cast(data);
-    if (p->desktop == number) {
+    if (p->desktop == static_cast<quint32>(number)) {
         return;
     }
     p->desktop = number;
@@ -343,6 +430,7 @@ void PlasmaWindow::Private::unmappedCallback(void *data, org_kde_plasma_window *
 {
     auto p = cast(data);
     Q_UNUSED(window);
+    p->unmapped = true;
     emit p->q->unmapped();
     p->q->deleteLater();
 }
@@ -364,14 +452,24 @@ void PlasmaWindow::Private::stateChangedCallback(void *data, org_kde_plasma_wind
     p->setMaximizeable(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_MAXIMIZABLE);
     p->setMinimizeable(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_MINIMIZABLE);
     p->setSkipTaskbar(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_SKIPTASKBAR);
+    p->setShadeable(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_SHADEABLE);
+    p->setShaded(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_SHADED);
+    p->setMovable(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_MOVABLE);
+    p->setResizable(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_RESIZABLE);
+    p->setVirtualDesktopChangeable(state & ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_VIRTUAL_DESKTOP_CHANGEABLE);
 }
 
 void PlasmaWindow::Private::themedIconNameChangedCallback(void *data, org_kde_plasma_window *window, const char *name)
 {
     auto p = cast(data);
     Q_UNUSED(window);
-    QIcon icon = QIcon::fromTheme(QString::fromUtf8(name));
-    p->icon = icon;
+    const QString themedName = QString::fromUtf8(name);
+    if (!themedName.isEmpty()) {
+        QIcon icon = QIcon::fromTheme(themedName);
+        p->icon = icon;
+    } else {
+        p->icon = QIcon();
+    }
     emit p->q->iconChanged();
 }
 
@@ -492,6 +590,51 @@ void PlasmaWindow::Private::setSkipTaskbar(bool skip)
     emit q->skipTaskbarChanged();
 }
 
+void PlasmaWindow::Private::setShadeable(bool set)
+{
+    if (shadeable == set) {
+        return;
+    }
+    shadeable = set;
+    emit q->shadeableChanged();
+}
+
+void PlasmaWindow::Private::setShaded(bool set)
+{
+    if (shaded == set) {
+        return;
+    }
+    shaded = set;
+    emit q->shadedChanged();
+}
+
+void PlasmaWindow::Private::setMovable(bool set)
+{
+    if (movable == set) {
+        return;
+    }
+    movable = set;
+    emit q->movableChanged();
+}
+
+void PlasmaWindow::Private::setResizable(bool set)
+{
+    if (resizable == set) {
+        return;
+    }
+    resizable = set;
+    emit q->resizableChanged();
+}
+
+void PlasmaWindow::Private::setVirtualDesktopChangeable(bool set)
+{
+    if (virtualDesktopChangeable == set) {
+        return;
+    }
+    virtualDesktopChangeable = set;
+    emit q->virtualDesktopChangeableChanged();
+}
+
 PlasmaWindow::Private::Private(org_kde_plasma_window *w, quint32 internalId, PlasmaWindow *q)
     : internalId(internalId)
     , q(q)
@@ -622,6 +765,31 @@ QIcon PlasmaWindow::icon() const
     return d->icon;
 }
 
+bool PlasmaWindow::isShadeable() const
+{
+    return d->shadeable;
+}
+
+bool PlasmaWindow::isShaded() const
+{
+    return d->shaded;
+}
+
+bool PlasmaWindow::isResizable() const
+{
+    return d->resizable;
+}
+
+bool PlasmaWindow::isMovable() const
+{
+    return d->movable;
+}
+
+bool PlasmaWindow::isVirtualDesktopChangeable() const
+{
+    return d->virtualDesktopChangeable;
+}
+
 void PlasmaWindow::requestActivate()
 {
     org_kde_plasma_window_set_state(d->window,
@@ -634,6 +802,16 @@ void PlasmaWindow::requestClose()
     org_kde_plasma_window_close(d->window);
 }
 
+void PlasmaWindow::requestMove()
+{
+    org_kde_plasma_window_request_move(d->window);
+}
+
+void PlasmaWindow::requestResize()
+{
+    org_kde_plasma_window_request_resize(d->window);
+}
+
 void PlasmaWindow::requestVirtualDesktop(quint32 desktop)
 {
     org_kde_plasma_window_set_virtual_desktop(d->window, desktop);
@@ -675,10 +853,33 @@ void PlasmaWindow::unsetMinimizedGeometry(Surface *panel)
     org_kde_plasma_window_unset_minimized_geometry(d->window, *panel);
 }
 
+void PlasmaWindow::requestToggleShaded()
+{
+    if (d->shaded) {
+        org_kde_plasma_window_set_state(d->window,
+            ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_SHADED,
+            0);
+    } else {
+        org_kde_plasma_window_set_state(d->window,
+            ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_SHADED,
+            ORG_KDE_PLASMA_WINDOW_MANAGEMENT_STATE_SHADED);
+    }
+}
+
 quint32 PlasmaWindow::internalId() const
 {
     return d->internalId;
 }
 
+QPointer<PlasmaWindow> PlasmaWindow::parentWindow() const
+{
+    return d->parentWindow;
+}
+
+QRect PlasmaWindow::geometry() const
+{
+    return d->geometry;
+}
+
 }
 }
