@@ -23,14 +23,20 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "datadevice_interface.h"
 #include "datasource_interface.h"
 #include "keyboard_interface.h"
+#include "keyboard_interface_p.h"
 #include "pointer_interface.h"
+#include "pointer_interface_p.h"
 #include "surface_interface.h"
+#include "textinput_interface_p.h"
 // Wayland
 #ifndef WL_SEAT_NAME_SINCE_VERSION
 #define WL_SEAT_NAME_SINCE_VERSION 2
 #endif
 // linux
+#include <config-kwayland.h>
+#if HAVE_LINUX_INPUT_H
 #include <linux/input.h>
+#endif
 
 namespace KWayland
 {
@@ -202,25 +208,30 @@ DataDeviceInterface *SeatInterface::Private::dataDeviceForSurface(SurfaceInterfa
     return interfaceForSurface(surface, dataDevices);
 }
 
+TextInputInterface *SeatInterface::Private::textInputForSurface(SurfaceInterface *surface) const
+{
+    return interfaceForSurface(surface, textInputs);
+}
+
 void SeatInterface::Private::registerDataDevice(DataDeviceInterface *dataDevice)
 {
     Q_ASSERT(dataDevice->seat() == q);
     dataDevices << dataDevice;
-    QObject::connect(dataDevice, &QObject::destroyed, q,
-        [this, dataDevice] {
-            dataDevices.removeAt(dataDevices.indexOf(dataDevice));
-            if (keys.focus.selection == dataDevice) {
-                keys.focus.selection = nullptr;
-            }
-            if (currentSelection == dataDevice) {
-                // current selection is cleared
-                currentSelection = nullptr;
-                if (keys.focus.selection) {
-                    keys.focus.selection->sendClearSelection();
-                }
+    auto dataDeviceCleanup = [this, dataDevice] {
+        dataDevices.removeOne(dataDevice);
+        if (keys.focus.selection == dataDevice) {
+            keys.focus.selection = nullptr;
+        }
+        if (currentSelection == dataDevice) {
+            // current selection is cleared
+            currentSelection = nullptr;
+            if (keys.focus.selection) {
+                keys.focus.selection->sendClearSelection();
             }
         }
-    );
+    };
+    QObject::connect(dataDevice, &QObject::destroyed, q, dataDeviceCleanup);
+    QObject::connect(dataDevice, &Resource::unbound, q, dataDeviceCleanup);
     QObject::connect(dataDevice, &DataDeviceInterface::selectionChanged, q,
         [this, dataDevice] {
             updateSelection(dataDevice, true);
@@ -267,6 +278,33 @@ void SeatInterface::Private::registerDataDevice(DataDeviceInterface *dataDevice)
     }
 }
 
+
+void SeatInterface::Private::registerTextInput(TextInputInterface *ti)
+{
+    // text input version 0 might call this multiple times
+    if (textInputs.contains(ti)) {
+        return;
+    }
+    textInputs << ti;
+    if (textInput.focus.surface && textInput.focus.surface->client() == ti->client()) {
+        // this is a text input for the currently focused text input surface
+        if (!textInput.focus.textInput) {
+            textInput.focus.textInput = ti;
+            ti->d_func()->sendEnter(textInput.focus.surface, textInput.focus.serial);
+            emit q->focusedTextInputChanged();
+        }
+    }
+    QObject::connect(ti, &QObject::destroyed, q,
+        [this, ti] {
+            textInputs.removeAt(textInputs.indexOf(ti));
+            if (textInput.focus.textInput == ti) {
+                textInput.focus.textInput = nullptr;
+                emit q->focusedTextInputChanged();
+            }
+        }
+    );
+}
+
 void SeatInterface::Private::endDrag(quint32 serial)
 {
     auto target = drag.target;
@@ -293,6 +331,7 @@ void SeatInterface::Private::updateSelection(DataDeviceInterface *dataDevice, bo
                 keys.focus.selection->sendSelection(dataDevice);
             } else {
                 keys.focus.selection->sendClearSelection();
+                currentSelection = nullptr;
             }
         }
     }
@@ -347,14 +386,15 @@ void SeatInterface::Private::getPointer(wl_client *client, wl_resource *resource
 {
     // TODO: only create if seat has pointer?
     PointerInterface *pointer = new PointerInterface(q, resource);
-    pointer->create(display->getConnection(client), qMin(wl_resource_get_version(resource), s_pointerVersion), id);
+    auto clientConnection = display->getConnection(client);
+    pointer->create(clientConnection, qMin(wl_resource_get_version(resource), s_pointerVersion), id);
     if (!pointer->resource()) {
         wl_resource_post_no_memory(resource);
         delete pointer;
         return;
     }
     pointers << pointer;
-    if (globalPointer.focus.surface && globalPointer.focus.surface->client()->client() == client) {
+    if (globalPointer.focus.surface && globalPointer.focus.surface->client() == clientConnection) {
         // this is a pointer for the currently focused pointer surface
         if (!globalPointer.focus.pointer) {
             globalPointer.focus.pointer = pointer;
@@ -383,7 +423,8 @@ void SeatInterface::Private::getKeyboard(wl_client *client, wl_resource *resourc
 {
     // TODO: only create if seat has keyboard?
     KeyboardInterface *keyboard = new KeyboardInterface(q, resource);
-    keyboard->create(display->getConnection(client), qMin(wl_resource_get_version(resource), s_keyboardVersion) , id);
+    auto clientConnection = display->getConnection(client);
+    keyboard->create(clientConnection, qMin(wl_resource_get_version(resource), s_keyboardVersion) , id);
     if (!keyboard->resource()) {
         wl_resource_post_no_memory(resource);
         delete keyboard;
@@ -394,7 +435,7 @@ void SeatInterface::Private::getKeyboard(wl_client *client, wl_resource *resourc
         keyboard->setKeymap(keys.keymap.fd, keys.keymap.size);
     }
     keyboards << keyboard;
-    if (keys.focus.surface && keys.focus.surface->client()->client() == client) {
+    if (keys.focus.surface && keys.focus.surface->client() == clientConnection) {
         // this is a keyboard for the currently focused keyboard surface
         if (!keys.focus.keyboard) {
             keys.focus.keyboard = keyboard;
@@ -421,14 +462,15 @@ void SeatInterface::Private::getTouch(wl_client *client, wl_resource *resource,
 {
     // TODO: only create if seat has touch?
     TouchInterface *touch = new TouchInterface(q, resource);
-    touch->create(display->getConnection(client), qMin(wl_resource_get_version(resource), s_touchVersion), id);
+    auto clientConnection = display->getConnection(client);
+    touch->create(clientConnection, qMin(wl_resource_get_version(resource), s_touchVersion), id);
     if (!touch->resource()) {
         wl_resource_post_no_memory(resource);
         delete touch;
         return;
     }
     touchs << touch;
-    if (touchInterface.focus.surface && touchInterface.focus.surface->client()->client() == client) {
+    if (touchInterface.focus.surface && touchInterface.focus.surface->client() == clientConnection) {
         // this is a touch for the currently focused touch surface
         if (!touchInterface.focus.touch) {
             touchInterface.focus.touch = touch;
@@ -643,6 +685,7 @@ QMatrix4x4 SeatInterface::focusedPointerSurfaceTransformation() const
 namespace {
 static quint32 qtToWaylandButton(Qt::MouseButton button)
 {
+#if HAVE_LINUX_INPUT_H
     static const QHash<Qt::MouseButton, quint32> s_buttons({
         {Qt::LeftButton, BTN_LEFT},
         {Qt::RightButton, BTN_RIGHT},
@@ -663,6 +706,9 @@ static quint32 qtToWaylandButton(Qt::MouseButton button)
         // further mapping not possible, 0x120 is BTN_JOYSTICK
     });
     return s_buttons.value(button, 0);
+#else
+    return 0;
+#endif
 }
 }
 
@@ -714,6 +760,10 @@ void SeatInterface::pointerButtonPressed(quint32 button)
     }
     if (d->globalPointer.focus.pointer && d->globalPointer.focus.surface) {
         d->globalPointer.focus.pointer->buttonPressed(button, serial);
+        if (d->globalPointer.focus.surface == d->keys.focus.surface && d->keys.focus.keyboard) {
+            // update the focused child surface
+            d->keys.focus.keyboard->d_func()->focusChildSurface(d->globalPointer.focus.pointer->d_func()->focusedChildSurface, serial);
+        }
     }
 }
 
@@ -817,13 +867,18 @@ void SeatInterface::setFocusedKeyboardSurface(SurfaceInterface *surface)
         if (d->keys.focus.selection) {
             if (d->currentSelection) {
                 d->keys.focus.selection->sendSelection(d->currentSelection);
+            } else {
+                d->keys.focus.selection->sendClearSelection();
             }
         }
     }
-    if (!k) {
-        return;
+    if (k) {
+        k->setFocusedSurface(surface, serial);
+    }
+    // focused text input surface follows keyboard
+    if (hasKeyboard()) {
+        setFocusedTextInputSurface(surface);
     }
-    k->setFocusedSurface(surface, serial);
 }
 
 void SeatInterface::setKeymap(int fd, quint32 size)
@@ -1027,6 +1082,7 @@ qint32 SeatInterface::touchDown(const QPointF &globalPosition)
     if (d->touchInterface.focus.touch && d->touchInterface.focus.surface) {
         d->touchInterface.focus.touch->down(id, serial, globalPosition - d->touchInterface.focus.offset);
     } else if (id == 0 && focusedTouchSurface()) {
+#if HAVE_LINUX_INPUT_H
         auto p = d->pointerForSurface(focusedTouchSurface());
         if (!p) {
             return id;
@@ -1039,6 +1095,7 @@ qint32 SeatInterface::touchDown(const QPointF &globalPosition)
                                 wl_fixed_from_double(pos.x()), wl_fixed_from_double(pos.y()));
 
         wl_pointer_send_button(p->resource(), serial, timestamp(), BTN_LEFT, WL_POINTER_BUTTON_STATE_PRESSED);
+#endif
     }
 
     d->touchInterface.ids << id;
@@ -1069,6 +1126,7 @@ void SeatInterface::touchUp(qint32 id)
     if (d->touchInterface.focus.touch && d->touchInterface.focus.surface) {
         d->touchInterface.focus.touch->up(id, display()->nextSerial());
     } else if (id == 0 && focusedTouchSurface()) {
+#if HAVE_LINUX_INPUT_H
         const quint32 serial = display()->nextSerial();
         auto p = d->pointerForSurface(focusedTouchSurface());
         if (!p) {
@@ -1076,6 +1134,7 @@ void SeatInterface::touchUp(qint32 id)
         }
 
         wl_pointer_send_button(p->resource(), serial, timestamp(), BTN_LEFT, WL_POINTER_BUTTON_STATE_RELEASED);
+#endif
     }
     d->touchInterface.ids.removeAll(id);
 }
@@ -1146,5 +1205,75 @@ DataDeviceInterface *SeatInterface::dragSource() const
     return d->drag.source;
 }
 
+void SeatInterface::setFocusedTextInputSurface(SurfaceInterface *surface)
+{
+    Q_D();
+    const quint32 serial = d->display->nextSerial();
+    const auto old = d->textInput.focus.textInput;
+    if (d->textInput.focus.textInput) {
+        // TODO: setFocusedSurface like in other interfaces
+        d->textInput.focus.textInput->d_func()->sendLeave(serial, d->textInput.focus.surface);
+    }
+    if (d->textInput.focus.surface) {
+        disconnect(d->textInput.focus.destroyConnection);
+    }
+    d->textInput.focus = Private::TextInput::Focus();
+    d->textInput.focus.surface = surface;
+    TextInputInterface *t = d->textInputForSurface(surface);
+    if (t && !t->resource()) {
+        t = nullptr;
+    }
+    d->textInput.focus.textInput = t;
+    if (d->textInput.focus.surface) {
+        d->textInput.focus.destroyConnection = connect(surface, &QObject::destroyed, this,
+            [this] {
+                setFocusedTextInputSurface(nullptr);
+            }
+        );
+        d->textInput.focus.serial = serial;
+    }
+    if (t) {
+        // TODO: setFocusedSurface like in other interfaces
+        t->d_func()->sendEnter(surface, serial);
+    }
+    if (old != t) {
+        emit focusedTextInputChanged();
+    }
+}
+
+SurfaceInterface *SeatInterface::focusedTextInputSurface() const
+{
+    Q_D();
+    return d->textInput.focus.surface;
+}
+
+TextInputInterface *SeatInterface::focusedTextInput() const
+{
+    Q_D();
+    return d->textInput.focus.textInput;
+}
+
+DataDeviceInterface *SeatInterface::selection() const
+{
+    Q_D();
+    return d->currentSelection;
+}
+
+void SeatInterface::setSelection(DataDeviceInterface *dataDevice)
+{
+    Q_D();
+    if (d->currentSelection == dataDevice) {
+        return;
+    }
+    d->currentSelection = dataDevice;
+    if (d->keys.focus.selection) {
+        if (dataDevice) {
+            d->keys.focus.selection->sendSelection(dataDevice);
+        } else {
+            d->keys.focus.selection->sendClearSelection();
+        }
+    }
+}
+
 }
 }
