@@ -25,6 +25,8 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "region_interface.h"
 #include "subcompositor_interface.h"
 #include "subsurface_interface_p.h"
+// Qt
+#include <QListIterator>
 // Wayland
 #include <wayland-server.h>
 // std
@@ -45,14 +47,34 @@ SurfaceInterface::Private::~Private()
     destroy();
 }
 
-void SurfaceInterface::Private::addChild(QPointer< SubSurfaceInterface > subSurface)
+void SurfaceInterface::Private::addChild(QPointer< SubSurfaceInterface > child)
 {
-    pending.children.append(subSurface);
+    // protocol is not precise on how to handle the addition of new sub surfaces
+    pending.children.append(child);
+    subSurfacePending.children.append(child);
+    current.children.append(child);
+    Q_Q(SurfaceInterface);
+    emit q->subSurfaceTreeChanged();
+    QObject::connect(child.data(), &SubSurfaceInterface::positionChanged, q, &SurfaceInterface::subSurfaceTreeChanged);
+    QObject::connect(child->surface().data(), &SurfaceInterface::damaged, q, &SurfaceInterface::subSurfaceTreeChanged);
+    QObject::connect(child->surface().data(), &SurfaceInterface::unmapped, q, &SurfaceInterface::subSurfaceTreeChanged);
+    QObject::connect(child->surface().data(), &SurfaceInterface::subSurfaceTreeChanged, q, &SurfaceInterface::subSurfaceTreeChanged);
 }
 
-void SurfaceInterface::Private::removeChild(QPointer< SubSurfaceInterface > subSurface)
+void SurfaceInterface::Private::removeChild(QPointer< SubSurfaceInterface > child)
 {
-    pending.children.removeAll(subSurface);
+    // protocol is not precise on how to handle the addition of new sub surfaces
+    pending.children.removeAll(child);
+    subSurfacePending.children.removeAll(child);
+    current.children.removeAll(child);
+    Q_Q(SurfaceInterface);
+    emit q->subSurfaceTreeChanged();
+    QObject::disconnect(child.data(), &SubSurfaceInterface::positionChanged, q, &SurfaceInterface::subSurfaceTreeChanged);
+    if (!child->surface().isNull()) {
+        QObject::disconnect(child->surface().data(), &SurfaceInterface::damaged, q, &SurfaceInterface::subSurfaceTreeChanged);
+        QObject::disconnect(child->surface().data(), &SurfaceInterface::unmapped, q, &SurfaceInterface::subSurfaceTreeChanged);
+        QObject::disconnect(child->surface().data(), &SurfaceInterface::subSurfaceTreeChanged, q, &SurfaceInterface::subSurfaceTreeChanged);
+    }
 }
 
 bool SurfaceInterface::Private::raiseChild(QPointer<SubSurfaceInterface> subsurface, SurfaceInterface *sibling)
@@ -70,6 +92,7 @@ bool SurfaceInterface::Private::raiseChild(QPointer<SubSurfaceInterface> subsurf
         // it's to the parent, so needs to become last item
         pending.children.append(*it);
         pending.children.erase(it);
+        pending.childrenChanged = true;
         return true;
     }
     if (!sibling->subSurface()) {
@@ -86,6 +109,7 @@ bool SurfaceInterface::Private::raiseChild(QPointer<SubSurfaceInterface> subsurf
     // find the iterator again
     siblingIt = std::find(pending.children.begin(), pending.children.end(), sibling->subSurface());
     pending.children.insert(++siblingIt, value);
+    pending.childrenChanged = true;
     return true;
 }
 
@@ -105,6 +129,7 @@ bool SurfaceInterface::Private::lowerChild(QPointer<SubSurfaceInterface> subsurf
         auto value = *it;
         pending.children.erase(it);
         pending.children.prepend(value);
+        pending.childrenChanged = true;
         return true;
     }
     if (!sibling->subSurface()) {
@@ -121,6 +146,7 @@ bool SurfaceInterface::Private::lowerChild(QPointer<SubSurfaceInterface> subsurf
     // find the iterator again
     siblingIt = std::find(pending.children.begin(), pending.children.end(), sibling->subSurface());
     pending.children.insert(siblingIt, value);
+    pending.childrenChanged = true;
     return true;
 }
 
@@ -150,7 +176,7 @@ void SurfaceInterface::Private::setContrast(const QPointer<ContrastInterface> &c
 
 #ifndef DOXYGEN_SHOULD_SKIP_THIS
 const struct wl_surface_interface SurfaceInterface::Private::s_interface = {
-    destroyCallback,
+    resourceDestroyedCallback,
     attachCallback,
     damageCallback,
     frameCallaback,
@@ -179,6 +205,13 @@ void SurfaceInterface::frameRendered(quint32 msec)
         wl_callback_send_done(r, msec);
         wl_resource_destroy(r);
     }
+    for (auto it = d->current.children.constBegin(); it != d->current.children.constEnd(); ++it) {
+        const auto &subSurface = *it;
+        if (subSurface.isNull() || subSurface->d_func()->surface.isNull()) {
+            continue;
+        }
+        subSurface->d_func()->surface->frameRendered(msec);
+    }
     if (needsFlush)  {
         client()->flush();
     }
@@ -186,111 +219,150 @@ void SurfaceInterface::frameRendered(quint32 msec)
 
 void SurfaceInterface::Private::destroy()
 {
-    for (wl_resource *c : current.callbacks) {
-        wl_resource_destroy(c);
-    }
-    for (wl_resource *c : pending.callbacks) {
-        wl_resource_destroy(c);
+    // copy all existing callbacks to new list and clear existing lists
+    // the wl_resource_destroy on the callback resource goes into destroyFrameCallback
+    // which would modify the list we are iterating on
+    QList<wl_resource *> callbacksToDestroy;
+    callbacksToDestroy << current.callbacks;
+    current.callbacks.clear();
+    callbacksToDestroy << pending.callbacks;
+    pending.callbacks.clear();
+    callbacksToDestroy << subSurfacePending.callbacks;
+    subSurfacePending.callbacks.clear();
+    for (auto it = callbacksToDestroy.constBegin(), end = callbacksToDestroy.constEnd(); it != end; it++) {
+        wl_resource_destroy(*it);
     }
     if (current.buffer) {
         current.buffer->unref();
     }
-    if (resource) {
-        wl_resource_destroy(resource);
-        resource = nullptr;
-    }
 }
 
-void SurfaceInterface::Private::commit()
+void SurfaceInterface::Private::swapStates(State *source, State *target, bool emitChanged)
 {
     Q_Q(SurfaceInterface);
-    const bool bufferChanged = pending.bufferIsSet;
-    const bool opaqueRegionChanged = pending.opaqueIsSet;
-    const bool inputRegionChanged = pending.inputIsSet;
-    const bool scaleFactorChanged = current.scale != pending.scale;
-    const bool transformFactorChanged = current.transform != pending.transform;
-    const bool shadowChanged = pending.shadowIsSet;
-    const bool blurChanged = pending.blurIsSet;
-    const bool contrastChanged = pending.contrastIsSet;
-    const bool slideChanged = pending.slideIsSet;
+    bool bufferChanged = source->bufferIsSet;
+    const bool opaqueRegionChanged = source->opaqueIsSet;
+    const bool inputRegionChanged = source->inputIsSet;
+    const bool scaleFactorChanged = source->scaleIsSet && (target->scale != source->scale);
+    const bool transformChanged = source->transformIsSet && (target->transform != source->transform);
+    const bool shadowChanged = source->shadowIsSet;
+    const bool blurChanged = source->blurIsSet;
+    const bool contrastChanged = source->contrastIsSet;
+    const bool slideChanged = source->slideIsSet;
+    const bool childrenChanged = source->childrenChanged;
     bool sizeChanged = false;
-    auto buffer = current.buffer;
+    auto buffer = target->buffer;
     if (bufferChanged) {
+        // TODO: is the reffing correct for subsurfaces?
         QSize oldSize;
-        if (current.buffer) {
-            oldSize = current.buffer->size();
-            current.buffer->unref();
-            QObject::disconnect(current.buffer, &BufferInterface::sizeChanged, q, &SurfaceInterface::sizeChanged);
+        if (target->buffer) {
+            oldSize = target->buffer->size();
+            if (emitChanged) {
+                target->buffer->unref();
+                QObject::disconnect(target->buffer, &BufferInterface::sizeChanged, q, &SurfaceInterface::sizeChanged);
+            } else {
+                delete target->buffer;
+                target->buffer = nullptr;
+            }
         }
-        if (pending.buffer) {
-            pending.buffer->ref();
-            QObject::connect(pending.buffer, &BufferInterface::sizeChanged, q, &SurfaceInterface::sizeChanged);
-            const QSize newSize = pending.buffer->size();
+        if (source->buffer) {
+            if (emitChanged) {
+                source->buffer->ref();
+                QObject::connect(source->buffer, &BufferInterface::sizeChanged, q, &SurfaceInterface::sizeChanged);
+            }
+            const QSize newSize = source->buffer->size();
             sizeChanged = newSize.isValid() && newSize != oldSize;
         }
-        buffer = pending.buffer;
-    }
-    auto shadow = current.shadow;
-    if (shadowChanged) {
-        shadow = pending.shadow;
-    }
-    auto blur = current.blur;
-    if (blurChanged) {
-        blur = pending.blur;
-    }
-    auto contrast = current.contrast;
-    if (contrastChanged) {
-        contrast = pending.contrast;
-    }
-    auto slide = current.slide;
-    if (slideChanged) {
-        slide = pending.slide;
-    }
-    QList<wl_resource*> callbacks = current.callbacks;
-    callbacks.append(pending.callbacks);
-    // copy values
-    current = pending;
-    current.buffer = buffer;
-    current.callbacks = callbacks;
-    current.shadow = shadow;
-    current.blur = blur;
-    current.contrast = contrast;
-    current.slide = slide;
-    pending = State{};
-    pending.children = current.children;
-    pending.input = current.input;
-    pending.inputIsInfinite = current.inputIsInfinite;
-    pending.opaque = current.opaque;
-    // commit all subSurfaces to apply position changes
-    for (auto it = current.children.constBegin(); it != current.children.constEnd(); ++it) {
-        if (!(*it)) {
-            continue;
+        if (!target->buffer && !source->buffer && emitChanged) {
+            // null buffer set on a not mapped surface, don't emit unmapped
+            bufferChanged = false;
         }
-        (*it)->d_func()->commit();
+        buffer = source->buffer;
     }
-    if (opaqueRegionChanged) {
-        emit q->opaqueChanged(current.opaque);
+    // copy values
+    if (bufferChanged) {
+        target->buffer = buffer;
+        target->damage = source->damage;
+        target->bufferIsSet = source->bufferIsSet;
+    }
+    if (childrenChanged) {
+        target->childrenChanged = source->childrenChanged;
+        target->children = source->children;
+    }
+    target->callbacks.append(source->callbacks);
+
+    if (shadowChanged) {
+        target->shadow = source->shadow;
+        target->shadowIsSet = true;
+    }
+    if (blurChanged) {
+        target->blur = source->blur;
+        target->blurIsSet = true;
+    }
+    if (contrastChanged) {
+        target->contrast = source->contrast;
+        target->contrastIsSet = true;
+    }
+    if (slideChanged) {
+        target->slide = source->slide;
+        target->slideIsSet = true;
     }
     if (inputRegionChanged) {
-        emit q->inputChanged(current.input);
+        target->input = source->input;
+        target->inputIsInfinite = source->inputIsInfinite;
+        target->inputIsSet = true;
+    }
+    if (opaqueRegionChanged) {
+        target->opaque = source->opaque;
+        target->opaqueIsSet = true;
     }
     if (scaleFactorChanged) {
-        emit q->scaleChanged(current.scale);
+        target->scale = source->scale;
+        target->scaleIsSet = true;
     }
-    if (transformFactorChanged) {
-        emit q->transformChanged(current.transform);
+    if (transformChanged) {
+        target->transform = source->transform;
+        target->transformIsSet = true;
     }
-    if (bufferChanged) {
-        if (!current.damage.isEmpty()) {
+
+    *source = State{};
+    source->children = target->children;
+    if (opaqueRegionChanged) {
+        emit q->opaqueChanged(target->opaque);
+    }
+    if (inputRegionChanged) {
+        emit q->inputChanged(target->input);
+    }
+    if (scaleFactorChanged) {
+        emit q->scaleChanged(target->scale);
+    }
+    if (transformChanged) {
+        emit q->transformChanged(target->transform);
+    }
+    if (bufferChanged && emitChanged) {
+        if (!target->damage.isEmpty()) {
             const QRegion windowRegion = QRegion(0, 0, q->size().width(), q->size().height());
             if (!windowRegion.isEmpty()) {
-                current.damage = windowRegion.intersected(current.damage);
+                target->damage = windowRegion.intersected(target->damage);
+                if (emitChanged) {
+                    subSurfaceIsMapped = true;
+                    trackedDamage = trackedDamage.united(target->damage);
+                    emit q->damaged(target->damage);
+                    // workaround for https://bugreports.qt.io/browse/QTBUG-52092
+                    // if the surface is a sub-surface, but the main surface is not yet mapped, fake frame rendered
+                    if (subSurface && !subSurface->mainSurface()->buffer()) {
+                        q->frameRendered(0);
+                    }
+                }
             }
-            emit q->damaged(current.damage);
-        } else if (!current.buffer) {
+        } else if (!target->buffer && emitChanged) {
+            subSurfaceIsMapped = false;
             emit q->unmapped();
         }
     }
+    if (!emitChanged) {
+        return;
+    }
     if (sizeChanged) {
         emit q->sizeChanged();
     }
@@ -306,6 +378,46 @@ void SurfaceInterface::Private::commit()
     if (slideChanged) {
         emit q->slideOnShowHideChanged();
     }
+    if (childrenChanged) {
+        emit q->subSurfaceTreeChanged();
+    }
+}
+
+void SurfaceInterface::Private::commit()
+{
+    if (!subSurface.isNull() && subSurface->isSynchronized()) {
+        swapStates(&pending, &subSurfacePending, false);
+    } else {
+        swapStates(&pending, &current, true);
+        if (!subSurface.isNull()) {
+            subSurface->d_func()->commit();
+        }
+        // commit all subSurfaces to apply position changes
+        // "The cached state is applied to the sub-surface immediately after the parent surface's state is applied"
+        for (auto it = current.children.constBegin(); it != current.children.constEnd(); ++it) {
+            const auto &subSurface = *it;
+            if (subSurface.isNull()) {
+                continue;
+            }
+            subSurface->d_func()->commit();
+        }
+    }
+}
+
+void SurfaceInterface::Private::commitSubSurface()
+{
+    if (subSurface.isNull() || !subSurface->isSynchronized()) {
+        return;
+    }
+    swapStates(&subSurfacePending, &current, true);
+    // "The cached state is applied to the sub-surface immediately after the parent surface's state is applied"
+    for (auto it = current.children.constBegin(); it != current.children.constEnd(); ++it) {
+        const auto &subSurface = *it;
+        if (subSurface.isNull() || !subSurface->isSynchronized()) {
+            continue;
+        }
+        subSurface->d_func()->commit();
+    }
 }
 
 void SurfaceInterface::Private::damage(const QRect &rect)
@@ -320,6 +432,7 @@ void SurfaceInterface::Private::damage(const QRect &rect)
 void SurfaceInterface::Private::setScale(qint32 scale)
 {
     pending.scale = scale;
+    pending.scaleIsSet = true;
 }
 
 void SurfaceInterface::Private::setTransform(OutputInterface::Transform transform)
@@ -358,6 +471,9 @@ void SurfaceInterface::Private::attachBuffer(wl_resource *buffer, const QPoint &
             if (pending.buffer == buffer) {
                 pending.buffer = nullptr;
             }
+            if (subSurfacePending.buffer == buffer) {
+                subSurfacePending.buffer = nullptr;
+            }
             if (current.buffer == buffer) {
                 current.buffer->unref();
                 current.buffer = nullptr;
@@ -371,12 +487,7 @@ void SurfaceInterface::Private::destroyFrameCallback(wl_resource *r)
     auto s = cast<Private>(r);
     s->current.callbacks.removeAll(r);
     s->pending.callbacks.removeAll(r);
-}
-
-void SurfaceInterface::Private::destroyCallback(wl_client *client, wl_resource *resource)
-{
-    Q_UNUSED(client)
-    wl_resource_destroy(resource);
+    s->subSurfacePending.callbacks.removeAll(r);
 }
 
 void SurfaceInterface::Private::attachCallback(wl_client *client, wl_resource *resource, wl_resource *buffer, int32_t sx, int32_t sy)
@@ -554,6 +665,55 @@ QPointer< SlideInterface > SurfaceInterface::slideOnShowHide() const
     return d->current.slide;
 }
 
+bool SurfaceInterface::isMapped() const
+{
+    Q_D();
+    if (d->subSurface) {
+        // from spec:
+        // "A sub-surface becomes mapped, when a non-NULL wl_buffer is applied and the parent surface is mapped."
+        return d->subSurfaceIsMapped && !d->subSurface->parentSurface().isNull() && d->subSurface->parentSurface()->isMapped();
+    }
+    return d->current.buffer != nullptr;
+}
+
+QRegion SurfaceInterface::trackedDamage() const
+{
+    Q_D();
+    return d->trackedDamage;
+}
+
+void SurfaceInterface::resetTrackedDamage()
+{
+    Q_D();
+    d->trackedDamage = QRegion();
+}
+
+SurfaceInterface *SurfaceInterface::surfaceAt(const QPointF &position)
+{
+    if (!isMapped()) {
+        return nullptr;
+    }
+    Q_D();
+    // go from top to bottom. Top most child is last in list
+    QListIterator<QPointer<SubSurfaceInterface>> it(d->current.children);
+    it.toBack();
+    while (it.hasPrevious()) {
+        const auto &current = it.previous();
+        auto surface = current->surface();
+        if (surface.isNull()) {
+            continue;
+        }
+        if (auto s = surface->surfaceAt(position - current->position())) {
+            return s;
+        }
+    }
+    // check whether the geometry contains the pos
+    if (!size().isEmpty() && QRectF(QPoint(0, 0), size()).contains(position)) {
+        return this;
+    }
+    return nullptr;
+}
+
 SurfaceInterface::Private *SurfaceInterface::d_func() const
 {
     return reinterpret_cast<Private*>(d.data());
